import logging
from collections import defaultdict

logger = logging.getLogger('kafka_rest.events')

class FlushReason(object):
    LENGTH = 'length'
    TIME = 'time'
    SHUTDOWN = 'shutdown'

class DropReason(object):
    NONRETRIABLE = 'nonretriable'
    PRIMARY_QUEUE_FULL = 'primary_queue_full'
    RETRY_QUEUE_FULL = 'retry_queue_full'
    MAX_RETRIES_EXCEEDED = 'max_retries_exceeded'

class EventRegistrar(object):
    """
    Handles registration of callable event handlers which are
    executed in sequence when events are emitted.
    """
    def __init__(self, debug=False):
        self.handlers = defaultdict(list)
        self.debug = debug

    def register(self, event, handler):
        self.handlers[event].append(handler)

    def emit(self, event, *args, **kwargs):
        if self.debug:
            logger.debug('Event: {0} Args: {1} Kwargs: {2}'.format(event, args, kwargs))
        for handler in self.handlers[event]:
            try:
                handler(*args, **kwargs)
            except Exception as e:
                if event != 'event_handler_exception':
                    self.emit('event_handler_exception', event, e)



import types
from collections import defaultdict

from mock import Mock

from kafka_rest.client import KafkaRESTClient

def _mock_emitter(obj, event, *args, **kwargs):
    obj.event_mocks[event](*args, **kwargs)

class MockClient(KafkaRESTClient):
    @classmethod
    def with_basic_setup(cls):
        instance = cls(None, None, shutdown_timeout_seconds=0)
        instance.event_mocks = defaultdict(lambda: Mock())
        instance.registrar.emit = types.MethodType(_mock_emitter, instance)
        return instance

    def reset_mocks(self):
        for mock in self.event_mocks.values():
            mock.reset_mock()

    def mock_for(self, event):
        return self.event_mocks[event]



import logging
import json
from functools import partial
import time
try:
    from queue import Full, Empty
except ImportError:
    from Queue import Full, Empty
from collections import namedtuple

from tornado.ioloop import IOLoop, PeriodicCallback
from tornado.httpclient import AsyncHTTPClient
from tornado.escape import json_decode

from .rest_proxy import request_for_batch, RETRIABLE_ERROR_CODES
from .message import Message
from .events import FlushReason, DropReason

logger = logging.getLogger('kafka_rest.producer')

class AsyncProducer(object):
    def __init__(self, client):
        self.client = client
        self.flush_timers = {}
        self.retry_timer = None
        self.http_client = AsyncHTTPClient(io_loop=self.client.io_loop,
                                           max_clients=self.client.http_max_clients)

    def _schedule_retry_periodically(self):
        logger.debug('Scheduling retry queue processing every {0} seconds'.format(self.client.retry_period_seconds))
        self.retry_timer = PeriodicCallback(self._start_retries,
                                            self.client.retry_period_seconds * 1000)
        self.retry_timer.start()

    def _message_batches_from_queue(self, queue):
        current_time = time.time()
        current_batch = []
        while not queue.empty():
            try:
                message = queue.get_nowait()
            except Empty:
                break
            # If this is the retry queue, stop gathering if the first prioritized
            # item in the queue isn't due for retry yet. If this is the first-send
            # queue, this shouldn't ever trigger because retry_after_time is 0
            if message.retry_after_time > current_time:
                break
            current_batch.append(message)
            if len(current_batch) >= self.client.flush_max_batch_size:
                yield current_batch
                current_batch = []
        if current_batch:
            yield current_batch

    def _start_retries(self):
        """Go through all the retry queues and schedule produce callbacks
        for all messages that are due to be retried."""
        logger.debug('Checking retry queues for events to retry')
        for topic, retry_queue in self.client.retry_queues.items():
            for batch in self._message_batches_from_queue(retry_queue):
                logger.debug('Retrying batch of size {0} for topic {1}'.format(len(batch), topic))
                self.client.registrar.emit('retry_batch', topic, batch)
                IOLoop.current().add_callback(self._send_batch_produce_request, topic, batch)

    def _reset_flush_timer(self, topic):
        if topic in self.flush_timers:
            logger.debug('Clearing flush timer for topic {0}'.format(topic))
            IOLoop.current().remove_timeout(self.flush_timers[topic])
        logger.debug('Scheduled new flush timer for topic {0} in {1} seconds'.format(topic,
                                                                                     self.client.flush_time_threshold_seconds))
        handle = IOLoop.current().call_later(self.client.flush_time_threshold_seconds,
                                             self._flush_topic, topic, FlushReason.TIME)
        self.flush_timers[topic] = handle

    def _send_batch_produce_request(self, topic, batch):
        if self.client.in_shutdown:
            connect_timeout = self.client.shutdown_timeout_seconds
            request_timeout = self.client.shutdown_timeout_seconds
        else:
            connect_timeout = self.client.connect_timeout_seconds
            request_timeout = self.client.request_timeout_seconds
        request = request_for_batch(self.client.host, self.client.port,
                                    connect_timeout, request_timeout,
                                    self.client.schema_cache, topic, batch)
        logger.info('Sending {0} events to topic {1}'.format(len(batch), topic))
        self.client.registrar.emit('send_request', topic, batch)
        self.http_client.fetch(request,
                               callback=partial(self._handle_produce_response, topic),
                               raise_error=False)

    def _queue_message_for_retry(self, topic, message):
        if message.can_retry(self.client):
            new_message = message.for_retry(self.client)
            try:
                self.client.retry_queues[topic].put_nowait(new_message)
            except Full:
                logger.critical('Retry queue full for topic {0}, message {1} cannot be retried'.format(topic, message))
                self.client.registrar.emit('drop_message', topic, message, DropReason.RETRY_QUEUE_FULL)
            else:
                logger.debug('Queued failed message {0} for retry in topic {1}'.format(new_message, topic))
                self.client.registrar.emit('retry_message', topic, new_message)
        else:
            logger.critical('Dropping failed message {0} for topic {1}, has exceeded maximum retries'.format(message, topic))
            self.client.registrar.emit('drop_message', topic, message, DropReason.MAX_RETRIES_EXCEEDED)

    def _handle_produce_success(self, topic, response, response_body):
        # Store schema IDs if we haven't already
        if not isinstance(self.client.schema_cache['value'][topic], int):
            logger.debug('Storing value schema ID of {0} for topic {1}'.format(response_body['value_schema_id'], topic))
            self.client.schema_cache['value'][topic] = response_body['value_schema_id']
        if response_body.get('key_schema_id') and not isinstance(self.client.schema_cache['key'].get(topic), int):
            logger.debug('Storing key schema ID of {0} for topic {1}'.format(response_body['key_schema_id'], topic))
            self.client.schema_cache['key'][topic] = response_body['key_schema_id']

        # Individual requests could still have failed, need to check
        # each response object's error code
        succeeded, failed = [], []
        for idx, offset in enumerate(response_body['offsets']):
            message = response.request._batch[idx]
            if offset.get('error_code') == 1: # Non-retriable Kafka exception
                failed.append((message, offset))
                logger.critical('Got non-retriable Kafka exception "{0}" for message {1}'.format(offset.get('message'),
                                                                                               response.request._batch[idx]))
                self.client.registrar.emit('drop_message', topic, message, DropReason.NONRETRIABLE)
            elif offset.get('error_code') == 2: # Retriable Kafka exception
                failed.append((message, offset))
                self._queue_message_for_retry(topic, message)
            else:
                succeeded.append((message, offset))

        logger.info('Successful produce response for topic {0}. Succeeded: {1} Failed: {2}'.format(topic,
                                                                                                   len(succeeded),
                                                                                                   len(failed)))
        logger.debug('Failed messages with offsets: {0}'.format(failed))
        self.client.registrar.emit('produce_success', topic, succeeded, failed)

    def _handle_produce_response(self, topic, response):
        # First, we check for a transport error
        if response.error:
            logger.error('Transport error submitting batch to topic {0}: {1}'.format(topic, response.error))
            self.client.registrar.emit('transport_error', topic, response.error)
            for message in response.request._batch:
                self._queue_message_for_retry(topic, message)
            return

        # We should have gotten a well-formed response back from the
        # proxy if we got this far
        response_body = json_decode(response.body)
        error_code, error_message = response_body.get('error_code'), response_body.get('message')

        if response.code == 200:
            self._handle_produce_success(topic, response, response_body)
        else: # We failed somehow, more information in the error code
            if error_code in RETRIABLE_ERROR_CODES:
                for message in response.request._batch:
                    self._queue_message_for_retry(topic, message)
            else: # Non-retriable failure of entire request
                for message in response.request._batch:
                    self.client.registrar.emit('drop_message', topic, message, DropReason.NONRETRIABLE)

    def _flush_topic(self, topic, reason):
        logger.debug('Flushing topic {0} (reason: {1})'.format(topic, reason))
        self.client.registrar.emit('flush_topic', topic, reason)
        queue = self.client.message_queues[topic]
        for batch in self._message_batches_from_queue(queue):
            IOLoop.current().add_callback(self._send_batch_produce_request, topic, batch)
        if not self.client.in_shutdown:
            self._reset_flush_timer(topic)

    def evaluate_queue(self, topic, queue):
        if queue.qsize() >= self.client.flush_length_threshold:
            self._flush_topic(topic, FlushReason.LENGTH)
        elif topic not in self.flush_timers:
            self._reset_flush_timer(topic)

    def start_shutdown(self):
        """Prevent the producer from firing off any additional requests
        as a result of timers, then schedule the remainder of the shutdown
        tasks to take place after giving in-flight requests some time
        to return."""
        # We need to take manual control of the event loop now, so
        # we stop the timers in order to not fight against them
        for topic in self.flush_timers:
            logger.debug('Shutdown: removing flush timer for topic {0}'.format(topic))
            IOLoop.current().remove_timeout(self.flush_timers[topic])

        # Last-ditch send attempts on remaining messages. These will use
        # shorter shutdown timeouts on the request in order to finish
        # by the time we invoke _finish_shutdown
        IOLoop.current().add_callback(self._start_retries)
        for topic, queue in self.client.message_queues.items():
            if not queue.empty():
                IOLoop.current().add_callback(self._flush_topic, topic, FlushReason.SHUTDOWN)

        logger.debug('Shutdown: waiting {0} seconds for in-flight requests to return'.format(self.client.shutdown_timeout_seconds))

        # We issue this step in a separate callback to get around a small timing issue
        # with sending out all these requests before shutdown. If you imagine that the
        # _flush_topic calls above take 0.1 seconds each to complete, if we simply
        # registered this call here before any of those calls did their 0.1 seconds
        # of work, we would actually invoke _finish_shutdown before the last request
        # made had the full length of time allotted to it to finish its request.
        IOLoop.current().add_callback(lambda: IOLoop.current().call_later(self.client.shutdown_timeout_seconds,
                                                                          self._finish_shutdown))

    def _finish_shutdown(self):
        # Anything not sent at this point is not going to make it out. We
        # fire off a specialized event in this case to give the
        # application code a chance to do something with this data all
        # at once.
        self.client.registrar.emit('shutdown', self.client.message_queues, self.client.retry_queues)
        IOLoop.current().stop()
        logger.debug('Shutdown: producer issued stop command to IOLoop')



import sys
from unittest import TestCase
try:
    from queue import PriorityQueue
except ImportError:
    from Queue import PriorityQueue

from mock import Mock, patch, call

from .fixtures import MockClient
from kafka_rest.message import Message

class TestProducer(TestCase):
    def setUp(self):
        self.client = MockClient.with_basic_setup()
        self.producer = self.client.producer
        self.callback_mock = Mock(wraps=self.client.io_loop.add_callback)
        self.client.io_loop.add_callback = self.callback_mock

        self.test_schema = {
            'type': 'record',
            'name': 'test_driver',
            'fields': [
                {'name': 'val', 'type': 'int'}
            ]
        }
        self.test_value = {'val': 1}

        self.client.schema_cache['value']['test_driver'] = 1
        self.client.schema_cache['key']['test_driver'] = 2

    def tearDown(self):
        if not self.client.in_shutdown:
            self.client.shutdown(block=True)

    def test_message_batches_from_queue_single_batch(self):
        m1 = Message('test_driver', {'val': 1}, None, None, 0, 1)
        m2 = Message('test_driver', {'val': 2}, None, None, 0, 1)
        queue = self.client.message_queues['test_driver']
        for m in [m1, m2]:
            queue.put(m)

        result = list(self.producer._message_batches_from_queue(queue))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], [m1, m2])

    def test_message_batches_from_queue_multiple_batches(self):
        self.client.flush_max_batch_size = 1
        m1 = Message('test_driver', {'val': 1}, None, None, 0, 1)
        m2 = Message('test_driver', {'val': 2}, None, None, 0, 1)
        queue = self.client.message_queues['test_driver']
        for m in [m1, m2]:
            queue.put(m)

        result = list(self.producer._message_batches_from_queue(queue))
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], [m1])
        self.assertEqual(result[1], [m2])

    def test_message_batches_from_retry_queue(self):
        m1 = Message('test_driver', {'val': 1}, None, None, sys.maxsize, 1)
        m2 = Message('test_driver', {'val': 2}, None, None, 0, 1)
        queue = self.client.retry_queues['test_driver']
        for m in [m1, m2]:
            queue.put(m)

        result = list(self.producer._message_batches_from_queue(queue))
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], [m2])

    @patch('kafka_rest.producer.request_for_batch')
    def test_send_batch_produce_request(self, fake_request):
        m1 = Message('test_driver', {'val': 1}, None, None, 0, 1)
        m2 = Message('test_driver', {'val': 2}, None, None, 0, 1)
        batch = [m1, m2]

        self.producer._send_batch_produce_request('test_driver', batch)
        call_args = fake_request.mock_calls[0][1]

        self.assertEqual(call_args[2], 60)
        self.assertEqual(call_args[3], 60)
        self.assertEqual(call_args[5], 'test_driver')
        self.assertEqual(call_args[6], batch)

    @patch('kafka_rest.producer.request_for_batch')
    def test_send_batch_produce_request_in_shutdown(self, fake_request):
        m1 = Message('test_driver', {'val': 1}, None, None, 0, 1)
        m2 = Message('test_driver', {'val': 2}, None, None, 0, 1)
        batch = [m1, m2]

        self.client.in_shutdown = True
        self.producer._send_batch_produce_request('test_driver', batch)
        call_args = fake_request.mock_calls[0][1]

        self.assertEqual(call_args[2], 0)
        self.assertEqual(call_args[3], 0)
        self.assertEqual(call_args[5], 'test_driver')
        for idx, m in enumerate(batch):
            self.assertTrue(m.true_equals(call_args[6][idx]))

    def test_queue_message_for_retry(self):
        m1 = Message('test_driver', {'val': 1}, None, None, 0, 1)
        self.producer._queue_message_for_retry('test_driver', m1)
        retry_message = self.client.retry_queues['test_driver'].get()

        self.assertEqual(retry_message.attempt_number,
                         m1.attempt_number + 1)
        self.assertNotEqual(retry_message.retry_after_time,
                            m1.retry_after_time)
        self.client.mock_for('retry_message').assert_called_once_with('test_driver',
                                                                      retry_message)

    def test_queue_message_for_retry_when_full(self):
        m1 = Message('test_driver', {'val': 1}, None, None, 0, 1)
        self.client.retry_queues['test_driver'] = PriorityQueue(maxsize=1)
        self.client.retry_queues['test_driver'].put('whatever')

        self.producer._queue_message_for_retry('test_driver', m1)
        self.client.mock_for('drop_message').assert_called_once_with('test_driver',
                                                                     m1,
                                                                     'retry_queue_full')

    def test_queue_message_for_retry_nonretriable(self):
        m1 = Message('test_driver', {'val': 1}, None, None, 0, 10)
        self.producer._queue_message_for_retry('test_driver', m1)
        self.client.mock_for('drop_message').assert_called_once_with('test_driver',
                                                                     m1,
                                                                     'max_retries_exceeded')

    def test_handle_produce_success_stores_value_schema_id(self):
        self.client.schema_cache['value']['test_driver'] = None
        body = {'offsets': [], 'value_schema_id': 5, 'key_schema_id': 2}

        self.producer._handle_produce_success('test_driver', None, body)

        self.assertEqual(self.client.schema_cache['value']['test_driver'], 5)
        self.assertEqual(self.client.schema_cache['key']['test_driver'], 2)

    def test_handle_produce_success_simple_success(self):
        m1 = Message('test_driver', {'val': 1}, None, None, 0, 1)
        response = Mock()
        response.request = Mock()
        response.request._batch = [m1]
        body = {'offsets': [{}], 'value_schema_id': 1, 'key_schema_id': 2}

        self.producer._handle_produce_success('test_driver', response, body)

        self.client.mock_for('produce_success').assert_called_once_with('test_driver',
                                                                        [(m1, {})],
                                                                        [])

    @patch('kafka_rest.producer.AsyncProducer._queue_message_for_retry')
    def test_handle_produce_success_mixed_success(self, fake_retry):
        m1 = Message('test_driver', {'val': 1}, None, None, 0, 1)
        m2 = Message('test_driver', {'val': 2}, None, None, 0, 1)
        m3 = Message('test_driver', {'val': 3}, None, None, 0, 1)
        offsets = [{},
                   {'error_code': 1, 'message': 'Nonretriable'},
                   {'error_code': 2, 'message': 'Retriable'}]
        response = Mock()
        response.request = Mock()
        response.request._batch = [m1, m2, m3]
        body = {'offsets': offsets,
                'value_schema_id': 1,
                'key_schema_id': 2}

        self.producer._handle_produce_success('test_driver', response, body)

        self.client.mock_for('produce_success').assert_called_once_with('test_driver',
                                                                        [(m1, {})],
                                                                        [(m2, offsets[1]),
                                                                         (m3, offsets[2])])
        self.client.mock_for('drop_message').assert_called_once_with('test_driver',
                                                                     m2,
                                                                     'nonretriable')
        fake_retry.assert_called_once_with('test_driver', m3)

    @patch('kafka_rest.producer.AsyncProducer._queue_message_for_retry')
    def test_handle_produce_response_transport_error(self, fake_retry):
        m1 = Message('test_driver', {'val': 1}, None, None, 0, 1)
        m2 = Message('test_driver', {'val': 2}, None, None, 0, 1)
        response = Mock()
        response.error = object()
        response.request = Mock()
        response.request._batch = [m1, m2]

        self.producer._handle_produce_response('test_driver', response)

        self.client.mock_for('transport_error').assert_called_once_with('test_driver',
                                                                        response.error)
        fake_retry.assert_has_calls([call('test_driver', m1),
                                     call('test_driver', m2)])

    @patch('kafka_rest.producer.AsyncProducer._handle_produce_success')
    @patch('kafka_rest.producer.json_decode')
    def test_handle_produce_response_on_200(self, fake_decode, fake_handle_success):
        m1 = Message('test_driver', {'val': 1}, None, None, 0, 1)
        m2 = Message('test_driver', {'val': 2}, None, None, 0, 1)
        response = Mock()
        response.code = 200
        response.error = False
        response.request = Mock()
        response.request._batch = [m1, m2]
        body = {'offsets': [{}, {}],
                'value_schema_id': 1,
                'key_schema_id': 2}
        fake_decode.return_value = body

        self.producer._handle_produce_response('test_driver', response)

        fake_handle_success.assert_called_once_with('test_driver',
                                                    response,
                                                    body)

    @patch('kafka_rest.producer.AsyncProducer._queue_message_for_retry')
    @patch('kafka_rest.producer.json_decode')
    def test_handle_produce_response_on_retriable(self, fake_decode, fake_retry):
        m1 = Message('test_driver', {'val': 1}, None, None, 0, 1)
        m2 = Message('test_driver', {'val': 2}, None, None, 0, 1)
        response = Mock()
        response.code = 500
        response.error = False
        response.request = Mock()
        response.request._batch = [m1, m2]
        body = {'offsets': [{}, {}],
                'value_schema_id': 1,
                'key_schema_id': 2,
                'error_code': 50001,
                'message': 'Retriable'}
        fake_decode.return_value = body

        self.producer._handle_produce_response('test_driver', response)

        fake_retry.assert_has_calls([call('test_driver', m1),
                                     call('test_driver', m2)])


    @patch('kafka_rest.producer.AsyncProducer._queue_message_for_retry')
    @patch('kafka_rest.producer.json_decode')
    def test_handle_produce_response_on_nonretriable(self, fake_decode, fake_retry):
        m1 = Message('test_driver', {'val': 1}, None, None, 0, 1)
        m2 = Message('test_driver', {'val': 2}, None, None, 0, 1)
        response = Mock()
        response.code = 500
        response.error = False
        response.request = Mock()
        response.request._batch = [m1, m2]
        body = {'offsets': [{}, {}],
                'value_schema_id': 1,
                'key_schema_id': 2,
                'error_code': 50101,
                'message': 'Nonretriable'}
        fake_decode.return_value = body

        self.producer._handle_produce_response('test_driver', response)

        calls = [call('test_driver', m1, 'nonretriable'),
                 call('test_driver', m2, 'nonretriable')]
        self.client.mock_for('drop_message').assert_has_calls(calls)



from tornado.httpclient import HTTPRequest
from tornado.escape import json_encode

ERROR_CODES = {
    40401: 'Topic not found',
    40402: 'Partition not found',
    50001: 'Zookeeper error',
    50002: 'Kafka error',
    50003: 'Retriable Kafka error',
    50101: 'Only SSL endpoints found, but SSL not supported for invoked API'
}

# Confluent didn't seem very confident that 50002 would
# always be non-retriable when I asked them about it,
# so we'll keep it as retriable for now to be safe.
# See https://github.com/confluentinc/kafka-rest/issues/147
RETRIABLE_ERROR_CODES = set([50001, 50002, 50003])

def request_for_batch(host, port, connect_timeout, request_timeout,
                      schema_cache, topic, batch):
    """Returns a Tornado HTTPRequest to the REST proxy
    representing the given message batch. This does not
    send the request, it just creates the object."""
    body = {'records': [{'value': message.value, 'key': message.key, 'partition': message.partition}
                        for message in batch]}

    # The REST proxy's API requires us to double-encode the schemas.
    # Don't ask why, because I have no idea.
    value_schema, key_schema = schema_cache['value'][topic], schema_cache['key'].get(topic)
    if isinstance(value_schema, dict):
        body['value_schema'] = json_encode(value_schema)
    else:
        body['value_schema_id'] = value_schema
    if key_schema:
        if isinstance(key_schema, dict):
            body['key_schema'] = json_encode(key_schema)
        else:
            body['key_schema_id'] = key_schema

    request = HTTPRequest('{0}:{1}/topics/{2}'.format(host, port, topic),
                          connect_timeout=connect_timeout,
                          request_timeout=request_timeout,
                          method='POST',
                          headers={'Accept': 'application/vnd.kafka.v1+json',
                                   'Content-Type': 'application/vnd.kafka.avro.v1+json'},
                          body=json_encode(body))

    # We also stick the message batch on the HTTPRequest object itself
    # so it is available to us when we handle the response. This is necessary
    # because individual messages can fail even if the overall request is
    # successful.
    request._batch = batch
    return request



from setuptools import setup, find_packages

from kafka_rest import __version__

setup(
    name='kafka-rest',
    version=__version__,
    description="Async client for Confluent's Kafka REST Proxy",
    url='https://github.com/gamechanger/kafka-rest',
    author='GameChanger',
    author_email='travis@gc.io',
    packages=find_packages(),
    install_requires=['tornado>=4.0.0,<5.0.0'],
    tests_require=['mock==1.3.0', 'nose==1.3.7'],
    test_suite="nose.collector",
    zip_safe=False
)
